#!/usr/bin/env python

"""
Usage: dictdefence [-dh] [-c <config file>] [-w <whitelist file>] 
        [-r <regex file>] [-e <command>]

The purpose of this software is prevent incoming dictionary based
attacks to a particular machine, or if run on a firewall, many machines.
It provides extensible methods of identifying attacks as well as
a variety of extensible methods of dealing with the attacks as they
happen.

This software is meant to run as a daemon and has four distinct
pieces, all of which handle different tasks.

- Readers:
    The readers, each of which run in their own thread, attach to a
    file or, preferably, a FIFO (named pipe) and continuously 
    analyze incoming data based upon user-defined perl-compatible
    regular expressions.  When a "bad" access is tripped, that
    reader notifies the database thread.
    
- Database Accessors:
    There is a single thread running which receives notifications of
    accesses from the reader threads.  It then logs this access in
    your chosen database and determines if your threshold for 
    failures have been exceeded.  If they have been exceeded, it logs
    a ban and notifies the firewall thread.
    
    The database thread also routinely runs an internal cleanup
    routine which removes old entries from the access table and 
    unbans IPs who's ban period has been exceeded
    
- "Firewall" Manipulator:
    This thread receives the notifies from the database thread requesting
    either a ban or an unban and executes that depending on the type
    of "Firewall" you are running.
    
- Main Listener:
    The main thread runs a unix socket server which handles command
    line requests of the "-e <command>" nature and returns results.
    
The options are as follows:

-h, --help
    Displays the usage on the command line.
    
-d, --daemonize
    This will fork the process into the background to run as a daemon.
    This is generally how this should be run unless debugging.
    
-c <config file>, --config=<config file>
    Specify the path to an alternate config file on the command 
    line.  The default config file is 
    /etc/dictdefence/dictdefence.conf.  The sample conf file
    contains documentation on all the options.
    
-w <whitelist file>, --whitelist=<whitelist file>
    Specify the path to an alternate whitelist file on the command
    line.  The default whitelist file is
    /etc/dictdefence/dictdefence.whitelist.  The whitelist file
    contains documentation on whitelisting IPs.
    
-r <regex file>, --regex=<regex file>
    Specify the path to an alternate regex file on the command
    line. The default regex file is
    /etc/dictdefence/dictdefence.regex.  The regex file contains
    default regular expressions to use and documentation on
    adding your own.
    
-e <command>, --exec <command>
    Speficy a command to be run against a running server.  This
    allows you to ban and unban IPs as well as view the 
    currently banned IPs, access attempts and previously banned
    IPs.  The command should be one of the following:
        
        ban <IP>
        unban <IP>
        show (accesses|banned|prev_banned)
        
    The <IP> arguments should be in dotted quad ipv4 format.
    The different choices for the "show" command will
    list everything currently in the accesses table, banned
    table or previously banned table respectively.
    

FUTURE WORK

- Remote execution:
    The next version of this will hopefully have the ability to
    access remote machines to execute rules on them, rather
    than having to run this on the firewall itself.  Ideally,
    the hope would be to run this on a central syslog server
    which can then contact the firewall machine and execute
    the necessary commands on it.
    
"""

__cvsversion__ = '$Id: dictdefence.py,v 1.16 2007/12/06 23:32:45 jay Exp $'
__author__ = 'Jay Deiman; jay@splitstreams.com'
__copyright__ = 'Copyright 2007 Jason Deiman'
__license__ = 'GPLv3'
__version__ = '0.1'

from dictdeflib import utils
import sys , dictdeflib , Queue , os , time , getopt , socket

DEBUG = False

def procCmdLine():
    """
    Process the command line and return an instance of 
    dictdeflib.Config() based upon what is passed on the
    command-line or run an RPC against a running server
    if the -e <command> syntax is used
    """
    shortOpts = 'hdc:w:r:e'
    longOpts = ['help' , 'daemonize' , 'config=' , 'whitelist=' ,
                'regex=' , 'exec']
    confFile = None
    whiteFile = None
    regexFile = None
    daemonize = False
    try:
        (optList , coms) = getopt.getopt(sys.argv[1:] , shortOpts, longOpts)
    except getopt.GetOptError , e:
        print e
        usage(1)
    rcoms = False
    for opt , val in optList:
        if opt == '-h' or opt == '--help':
            usage()
        elif opt == '-d' or opt == '--daemonize':
            daemonize = True
        elif opt == '-c' or opt == '--config':
            confFile = val
        elif opt == '-w' or opt == '--whitelist':
            whiteFile = val
        elif opt == '-r' or opt == '--regex':
            regexFile = val
        elif opt == '-e' or opt == '--exec':
            rcoms = True
    config = dictdeflib.Config(daemonize , confFile , whiteFile , regexFile)
    if rcoms:
        runComs(config , coms)
    return config
    
def usage(exitCode=0):
    """
    Prints usage and exits
    """
    print 'Usage: %s [-dh] [-c <config file>]' % os.path.basename(sys.argv[0]),
    print '[-w <whitelist file>] [-r <regex file>] [-e <command>]'
    print """
    -h,--help         What you are looking at
    -d,--daemonize    Run as a daemon
    -c,--config=      Path to the config file
    -w,--whitelist=   Path to a whitelist file
    -r,--regex=       Path to the regex file
    -e,--exec         Use this to just make requests against a running
                      server.  The possible requests are:
                          
                          ban <ip>
                          unban <ip>
                          show (accesses|banned|prev_banned)
                      
                      Ban and unban must be followed by an ipv4 address.
                      Show must be followed by one of the above options.
    """
    sys.exit(exitCode)
    
def runComs(config , coms):
    """
    This is essentially the method which runs the remote procedure
    calls against the socket
    """
    s = socket.socket(socket.AF_UNIX , socket.SOCK_STREAM)
    s.connect(config.SockPath)
    s.send(' '.join(coms))
    data = ''
    while True:
        temp = s.recv(1024)
        if not temp: break
        data += temp
    s.close()
    if data[-1:] == '\n':
        print data,
    else:
        print data
    sys.exit(0)
    
def writePid():
    """
    Writes the pid file out to /var/run/dictdefence.pid
    """
    try:
        fh = open('/var/run/dictdefence.pid' , 'w')
    except IOError , e:
        return False
    fh.write(str(os.getpid()))
    fh.close()
    return True

def main():
    """
    The main entry point of the program
    """
    global DEBUG
    utils.DEBUG = DEBUG
    # Init the config from the command line opts
    config = procCmdLine()
    if config.Daemonize:
        # Daemonize the process by forking
        utils.daemonize()
    # Set up the other utility objects based on the config
    logger = dictdeflib.Logger(config)
    reBls = dictdeflib.REBlocks(config)
    wl = dictdeflib.Whitelist(config , logger)
    # Write the PID file
    if not writePid():
        logger.log('Could not write pid file' , logger.LOG_ERR)
    # Set up the inter-thread communication queues
    actionQ = Queue.Queue()
    fwQ = Queue.Queue()
    dbReqQ = Queue.Queue()
    dbRespQ = Queue.Queue()
    # Set the DB
    db = config.CBDBType(config , actionQ , fwQ , dbReqQ , dbRespQ , logger)
    # Set the rules
    rules = config.CBRuleType(config , fwQ , logger)
    # Set up the reader
    readers = []
    for f in config.ReadFiles:    
        readers.append(dictdeflib.LogReader(f , logger , actionQ , reBls , wl))
    # Start all the threads
    if utils.DEBUG:
        print 'Starting threads'
    db.start()
    rules.start()
    for reader in readers:
        reader.start()
    # Start the unix sock server listening
    if os.path.exists(config.SockPath): os.unlink(config.SockPath)
    srvr = dictdeflib.MyThreadingUnixStreamServer(config.SockPath ,
                dictdeflib.ReqHandler , dbReqQ , dbRespQ)
    os.chmod(config.SockPath , 0600)
    srvr.serve_forever()
    
if __name__ == '__main__':
    main()
