#!/usr/bin/env python

"""
This is a set of classes to handle the starting of a Unix socket server
and the associated request handler.  These are overrides of builtin
SocketServer classes to add some custom args to the server and 
handler.
"""

__cvsversion__ = '$Id: ClServer.py,v 1.4 2007/12/06 23:32:20 jay Exp $'
__author__ = 'Jay Deiman; jay@splitstreams.com'
__copyright__ = 'Copyright 2007 Jason Deiman'
__license__ = 'GPLv3'
__all__ = ['MyThreadingUnixStreamServer' , 'ReqHandler']

import SocketServer , re , time , utils

class MyThreadingUnixStreamServer (SocketServer.ThreadingUnixStreamServer):
    """
    This is the extension of the ThreadingUnixStreamServer which will
    enable the use of extra varibles in the request handler
    """
    def __init__(self , server_address , RequestHandlerClass , reqQ , respQ):
        """
        Overriding the constructor to add a couple of class level variables.
        """
        self.reqQ = reqQ
        self.respQ = respQ
        SocketServer.ThreadingUnixStreamServer.__init__(self , 
                server_address , RequestHandlerClass)
        
class ReqHandler(SocketServer.BaseRequestHandler):
    """
    This is an override of the BaseRequestHandler that will define 
    the specific handler
    """
    validQuestions = ['ban' , 'unban' , 'show']
    validShows = ['accesses' , 'banned' , 'prev_banned']
    
    def handle(self):
        """
        Handles the actual request
        """
        question = self.request.recv(1024)
        # Split the question into words
        req = question.split(' ')
        if not req[0] in self.validQuestions:
            self._error('Invalid request type: %s' % req[0])
            return
        validIp = re.compile(r'(\d{1,3}\.){3}\d{1,3}')
        if req[0] == 'ban' or req[0] == 'unban':
            # Check for error conditions first
            if len(req) < 2:
                self._error('You must specify an IP to %s' % req[0])
                return
            if not validIp.match(req[1]):
                self._error('Invalid IP address: %s' % req[1])
                return
            # Now process what should be our valid request
            if req[0] == 'ban':
                self.server.reqQ.put(tuple(req))
                self.request.send('Banned %s' % req[1])
            else:
                self.server.reqQ.put(tuple(req))
                self.request.send('Unbanned %s' % req[1])
            self.request.close()
            return
        # If we get here, we have a show request. Process errors first
        if len(req) < 2:
            self._error('You must specify something to show: accesses' + 
                        ', banned or prev_banned')
            return
        if not req[1] in self.validShows:
            self._error('You must specify something valid to show: accesses' + 
                        ', banned or prev_banned')
            return
        # Now process our valid show request
        data = []
        if req[1] == 'accesses':
            self.server.reqQ.put('getAccesses')
            data = self.server.respQ.get()
            self.request.send('Accesses:\n')
        elif req[1] == 'banned':            
            self.server.reqQ.put('getBanned')
            data = self.server.respQ.get()
            self.request.send('Currently Banned:\n')
        else:
            self.server.reqQ.put('getPrevBanned')
            data = self.server.respQ.get()
            self.request.send('Previously Banned:\n')
        # Send the header
        self.request.send('%-27s%s\n' % ('Time' , 'IP'))
        self.request.send('%-27s%s\n' % ('-' * 24 , '-' * 15))
        # Now send the data
        for row in data:
            self.request.send('%-27s%s\n' % (time.ctime(row['ts']) , 
                                             row['ip_str']))
        self.request.close()
            
    def _error(self , msg):
        self.request.send('ERROR: %s' % msg)
        self.request.close()