#!/usr/bin/env python

"""
The class which defines the the methods for reading and updating
and MySQL database
"""

__cvsversion__ = '$Id: MySQLDB.py,v 1.14 2007/12/06 23:32:20 jay Exp $'
__author__ = 'Jay Deiman; jay@splitstreams.com'
__copyright__ = 'Copyright 2007 Jason Deiman'
__license__ = 'GPLv3'
__all__ = ['MySQLDB']

import utils , time , MySQLdb , DB

class MySQLDB(DB.DB):
    """
    Implements the DB class for MySQL
    """
    def __init__(self , config , actionQ , fwQ , reqQ , respQ , logger):
        DB.DB.__init__(self , config , actionQ , fwQ , reqQ , respQ , logger)
        
    def _cleanup(self):
        """
        Cleanup the database accesses table.  Also unban expired bans.
        """
        # First we remove anything in the accesses table older than
        # sixty days
        sixtyDaysAgo = time.time() - (60 * 60 * 24 * 60)
        delete = 'DELETE FROM accesses WHERE ts < %s'
        self._cursor.execute(delete , (sixtyDaysAgo,))
        self._conn.commit()
        # Now we unban any expired bans
        expireTime = time.time() - self._config.BlockTime
        select = 'SELECT ip_str FROM banned WHERE ts < %s'
        self._cursor.execute(select , (expireTime,))
        for row in self._cursor:
            self.unban(row[0])        
    
    def _initDB(self):
        """
        Initialize the connection to the database and create the
        database if it does not exist
        """
        conn = None
        try:
            if self._config.DBUnixSocket:
                conn = MySQLdb.connect(user=self._config.DBUser ,
                                         passwd=self._config.DBPass ,
                                         unix_socket=self._config.DBUnixSocket ,
                                         db=self._config.DBName)
            else:
                conn = MySQLdb.connect(user=self._config.DBUser ,
                                         passwd=self._config.DBPass ,
                                         db=self._config.DBName ,
                                         host=self._config.DBHost ,
                                         port=self._config.DBPort)
        except MySQLdb.MySQLError , e:
            self._logger.log(str(e) , self._logger.LOG_ERR)
            raise MySQLdb.MySQLError , e
        
        curs = conn.cursor()
        # Check to see if the tables already exist
        numres = curs.execute('SHOW TABLES')
        if numres == 3:
            curs.close()
            return conn
        # Or else we create the tables
        tables = ("""
            CREATE TABLE accesses (
                acc_id INTEGER UNSIGNED AUTO_INCREMENT PRIMARY KEY ,
                ts DOUBLE ,
                ip_int INTEGER UNSIGNED,
                ip_str VARCHAR(15) ,
                logline VARCHAR(255)
            );
            """ ,
            'CREATE INDEX ts_acc_idx ON accesses (ts);' ,
            'CREATE INDEX ip_int_acc_idx ON accesses (ip_int);' ,
            """
            CREATE TABLE banned (
                ban_id INTEGER UNSIGNED AUTO_INCREMENT PRIMARY KEY ,
                ip_int INTEGER UNSIGNED ,
                ts DOUBLE ,
                ip_str VARCHAR(15) ,
                permanent BIT
            );
            """ ,
            'CREATE INDEX ts_ban_idx ON banned (ts);' ,
            'CREATE INDEX ip_int_ban_idx ON banned (ip_int);' ,
            """
            CREATE TABLE prev_banned (
                ban_id INTEGER UNSIGNED AUTO_INCREMENT PRIMARY KEY ,
                ts DOUBLE ,
                ip_int INTEGER UNSIGNED ,
                ip_str VARCHAR(15)
            );
            """ ,
            'CREATE INDEX ip_int_pban_idx ON prev_banned (ip_int);' ,
        )
        curs = conn.cursor()
        # Create the tables
        for q in tables:
            curs.execute(q)
        curs.close()
        return conn
    
    def add(self , ip ,  logLine):
        """
        Adds an item to the database and passes the info to the
        firewall Queue if necessary
        """
        ipInt = utils.ip2int(ip)
        # Insert the access
        ins = (time.time() , ipInt , ip , logLine)
        insert = """
            INSERT INTO accesses (acc_id , ts , ip_int , ip_str , logline)
            VALUES (NULL , %s , %s , %s , %s)
        """
        self._cursor.execute(insert , ins)
        self._conn.commit()
        # Now check if this IP should be banned
        sel = (time.time() - self._config.TimePeriod , ipInt)
        select = 'SELECT COUNT(*) FROM accesses WHERE ts > %s AND ip_int = %s'
        self._cursor.execute(select , sel)
        count = self._cursor.fetchall()[0][0]
        if count >= self._config.NumFails:
            # Ban them
            self.ban(ip)
    
    def isBanned(self , ip):
        """
        Returns a boolean value stating whether the given IP is
        banned
        """
        ipInt = utils.ip2int(ip)
        select = 'SELECT COUNT(*) FROM banned WHERE ip_int = %s'
        self._cursor.execute(select , (ipInt,))
        if int(self._cursor.fetchall()[0][0]) > 0:
            return True
        return False
    
    def ban(self , ip):
        """
        This will ban a given IP by adding it to the database and
        queuing an entry for the firewall thread.  This will also
        check to see if the given IP needs to be permanently
        banned as well.
        """
        ipInt = utils.ip2int(ip)
        # Check to see if this has been banned before
        sel = (ipInt,)
        select = 'SELECT COUNT(*) FROM prev_banned WHERE ip_int = %s'
        self._cursor.execute(select , sel)
        count = int(self._cursor.fetchall()[0][0]) + 1
        if count >= self._config.NumFails:
            # We ban permanently if its over the numfails
            self.banPermanent(ip)
        else:
            # Check to see if the ip is already banned
            select = 'SELECT COUNT(*) FROM banned where ip_int = %s'
            self._cursor.execute(select , sel)
            if int(self._cursor.fetchall()[0][0]) > 0:
                self._logger.log('IP already banned: %s' % ip)
                return
            # Or else we just do a temporary ban
            ins = (ipInt , time.time() , ip)
            insert = """
                INSERT INTO banned (ip_int , ts , ip_str , permanent)
                VALUES (%s , %s , %s , 0)
            """
            self._cursor.execute(insert , ins)
            self._conn.commit()
            if utils.DEBUG:
                print 'Putting ban request on the queue'
            self._fwQ.put(('ban' , ip))
    
    def unban(self , ip):
        """
        This will unban an IP, moving the entry into the prev_banned
        table and putting an unban request in the firewall queue.
        """
        intIp = utils.ip2int(ip)
        # First delete from the banned table
        delete = 'DELETE FROM banned WHERE ip_int = %s'
        self._cursor.execute(delete , (intIp,))
        # Now insert into the previously banned table
        ins = (time.time() , intIp , ip)
        insert = """
            INSERT INTO prev_banned (ban_id , ts , ip_int , ip_str)
            VALUES (NULL , %s , %s , %s)
        """
        self._cursor.execute(insert , ins)
        self._conn.commit()
        # Add an unban to firewall queue
        self._fwQ.put(('unban' , ip))
    
    def banPermanent(self , ip):
        """
        Bans an IP with the permanent flag set
        """
        ins = (utils.ip2int(ip) , time.time() , ip)
        insert = """
            INSERT INTO banned (ip_int , ts , ip_str , permanent)
            VALUES (%s , %s , %s , 1)
        """
        self._cursor.execute(insert , ins)
        self._conn.commit()
        self._fwQ.put(('ban' , 'ip'))
    
    def getBanned(self):
        """
        Returns a list of hashes of data
        """
        data = []
        select = 'SELECT * FROM banned ORDER BY ts DESC'
        self._cursor.execute(select)
        colNames = [desc[0] for desc in self._cursor.description]
        for row in self._cursor:
            data.append(dict(zip(colNames , row)))
        return data
    
    def getAccesses(self):
        """
        Returns a list of hashes of data
        """
        data = []
        select = 'SELECT * FROM accesses ORDER BY ts DESC'
        self._cursor.execute(select)
        colNames = [desc[0] for desc in self._cursor.description]
        for row in self._cursor:
            data.append(dict(zip(colNames , row)))
        return data
    
    def getPrevBanned(self):
        """
        Returns a list of hashes of data
        """
        data = []
        select = 'SELECT * FROM prev_banned ORDER BY ts DESC'
        self._cursor.execute(select)
        colNames = [desc[0] for desc in self._cursor.description]
        for row in self._cursor:
            data.append(dict(zip(colNames , row)))
        return data
    
    def __del__(self):
        """
        Destructor: cleans up the database connections.
        """
        self._cursor.close()
        self._conn.close()
